#include "hardware.h"
#include "keyboard.h"


// ATMega32 timers:
// 8 bit timer0 - system ticks
// 16 bit timer1 -    1/10   
// 8 bit timer2 - beep (CW sidetone) -     PD7(OC2)

// AT91SAM7Sxxx timers
// TC0: UNUSED (or, start ADC sampling in test config).
// TC1: beep (CW sidetone) -   
// TC2: 1/10 dot length timer

static uint_fast32_t 
NOINLINEAT
calcdivround(
	uint_fast32_t freq		/*     ,  . */
	)
{
	return (CPU_FREQ + freq / 2) / freq;
}


 //       
static uint_fast8_t 
//NOINLINEAT
calcbits(
	unsigned long v		//   
	)
{
	uint_fast8_t n;

	for (n = 0; v != 0; ++ n)
		v >>= 1;

	return n;
}


//         .
//          .
//    ( )         (   1).
static uint_fast8_t 
//NOINLINEAT
calcdivider(
	unsigned long divider, //     
	uint_fast8_t width,			//    
	unsigned taps,			//  -  . 0x01 -  bypass, 0x02 -   2... 0x400 -   1024
	unsigned * dvalue)		//        (   1).
{
	const uint_fast8_t rbmax = calcbits(taps);
	uint_fast8_t rb, rbi;
	enum { substract = 1 };

	for (rb = rbi = 0; rb <= rbmax; ++ rb)
	{
		const unsigned prescaler = (1U << rb);
		if ((taps & prescaler) != 0)
		{
			//   .
			const unsigned long modulus = (divider >> rb) - substract;
			if (calcbits(modulus) <= width)
			{
				//   
				// rb -   -  . 
				// rbi -       .
				// modulus -       (   1).
				* dvalue = (unsigned) modulus;
				return rbi;
			}
			++ rbi;		//      .
		}
	}
	//          .
	* dvalue = 1 - substract;	//  
	return 0;	//      -  rbmax
}

	//   calcdivider().
	enum 
	{ 
		ATMEGA_SPCR_WIDTH = 0,		ATMEGA_SPCR_TAPS = (128 | 64 | 32 | 16 | 8 | 4 | 2), 
		ATMEGA128_TIMER0_WIDTH = 8,	ATMEGA128_TIMER0_TAPS = (1024 | 256 | 128 | 64 | 32 | 8 | 1), 
		ATMEGA_TIMER0_WIDTH = 8,	ATMEGA_TIMER0_TAPS = (1024 | 256 | 64 | 8 | 1), 
		ATMEGA_TIMER1_WIDTH = 16,	ATMEGA_TIMER1_TAPS = (1024 | 256 | 64 | 8 | 1),
		ATMEGA_TIMER2_WIDTH = 8,	ATMEGA_TIMER2_TAPS = (1024 | 256 | 128 | 64 | 32 | 8 | 1),	// 
		ATMEGA_UBR_WIDTH = 12,		ATMEGA_UBR_TAPS = (16 | 8),	/* UBR: USART Baud Rate Register */
		//ATMEGA_TWBR_WIDTH = 8,	ATMEGA_TWBR_TAPS = (64 | 16 | 4 | 1)	/* TWBR: TWI Bit Rate Register */
		ATMEGA_ADPS_WIDTH = 0,	ATMEGA_ADPS_TAPS = (128 | 64 | 32 | 16 | 8 | 4 | 2)	/* ADPS2:0: ADC Prescaler Select Bits */
	};
	//spcr_val = (1U << SPE) | (1U << MSTR) | (1U << CPOL) | (1U << CPHA) | (0U << SPR1) | (0U << SPR0);	// SPI MODE3, 0x5C
	//(1U << SPI2X)
	// spcr  spsr -  SPI
	static const PROGMEM struct { unsigned char spsr, spcr; } spcr_spsr [] =
	{
		{ (1U << SPI2X), 	(0U << SPR1) | (0U << SPR0), }, 	/* /2 */
		{ (0U << SPI2X), 	(0U << SPR1) | (0U << SPR0), }, 	/* /4 */
		{ (1U << SPI2X), 	(0U << SPR1) | (1U << SPR0), }, 	/* /8 */
		{ (0U << SPI2X), 	(0U << SPR1) | (1U << SPR0), }, 	/* /16 */
		{ (1U << SPI2X), 	(1U << SPR1) | (0U << SPR0), }, 	/* /32 */
		{ (0U << SPI2X), 	(1U << SPR1) | (0U << SPR0), }, 	/* /64 */
		{ (0U << SPI2X), 	(1U << SPR1) | (1U << SPR0), }, 	/* /128 */
	};
	// atmega128 timer0
	static const PROGMEM unsigned char m128_tccr0 [] = 
	{
		0x01, 	// 1 (no prescaler)
		0x02,	// 8
		0x03,	// 32
		0x04,	// 64
		0x05,	// 128
		0x06,	// 256
		0x07,	// 1024
	};
	// timer0  timer1
	//     PD7(OC2)
	//      .
	// 3  
	// ATMEGA32, ATMEGA644
	static const PROGMEM unsigned char tccr0_tccr1b [] = 
	{
		0x01, 	// 1 (no prescaler)
		0x02,	// 8
		0x03,	// 64
		0x04,	// 256
		0x05,	// 1024
	};
	// timer2 
	//     PD7(OC2)
	//      .
	// 3  
	static const PROGMEM unsigned char tccr2 [] = 
	{
		0x01,	// 1  (no prescaler)
		0x02,	// 8
		0x03,	// 32
		0x04,	// 64
		0x05,	// 128
		0x06,	// 256
		0x07,	// 1024
	};

	// ADC Control and Status Register A
	// 3  
	// ADPS2, ADPS1 and ADPS0 bits
	static const PROGMEM unsigned char adcsra [] = 
	{
		(1U << ADPS0),					// 2
		(1U << ADPS1),					// 4
		(1U << ADPS1) | (1U << ADPS0),	// 8
		(1U << ADPS2),					// 16
		(1U << ADPS2) | (1U << ADPS0),	// 32
		(1U << ADPS2) | (1U << ADPS1),	// 64
		(1U << ADPS2) | (1U << ADPS1) | (1U << ADPS0),	// 128
	};



	ISR(USART_RXC_vect)
	{
		cat_parsechar(UDR);
	}

	//ISR(USART_TXC_vect)
	//{
	//	cat_sendchar();
	//}

void hardware_cat_initialize(uint_fast8_t withirq)
{
#if CPUSTYLE_AT91SAM7
    // enable the clock of UART0
	AT91C_BASE_PMC->PMC_PCER = 1UL << AT91C_ID_US1;

	//AT91C_BASE_PIOA->PIO_MDER = AT91C_PA22_TXD1;	// 5 volt (open drain) output
    // disable I/O pullup
    AT91C_BASE_PIOA->PIO_PPUDR = AT91C_PA22_TXD1;
    // enable I/O pullup
    AT91C_BASE_PIOA->PIO_PPUER = AT91C_PA21_RXD1;

	// enable uart pins on PIO
    AT91C_BASE_PIOA->PIO_PDR = AT91C_PA21_RXD1 | AT91C_PA22_TXD1;	// enable peripheral control of PA21,PA22 (RXD1 and TXD1)
	AT91C_BASE_PIOA->PIO_ASR = AT91C_PA21_RXD1 | AT91C_PA22_TXD1; // assigns the 2 I/O lines to peripheral A function
   
	// reset the UART
    AT91C_BASE_US1->US_CR = AT91C_US_RSTRX | AT91C_US_RSTTX | AT91C_US_RXDIS | AT91C_US_TXDIS;
    // set serial line mode
    AT91C_BASE_US1->US_MR = 
						AT91C_US_OVER |
						AT91C_US_USMODE_NORMAL |// Normal Mode
                       AT91C_US_CLKS_CLOCK |   // Clock = MCK
                       AT91C_US_CHRL_8_BITS |
                       AT91C_US_PAR_NONE |
                       AT91C_US_NBSTOP_1_BIT;                   
	if (withirq)
	{
		AT91C_BASE_US1->US_IER = (AT91C_US_RXRDY | AT91C_US_TXRDY);

		// programming interrupts from ADC
		AT91C_BASE_AIC->AIC_IDCR = (1UL << AT91C_ID_US1);
		AT91C_BASE_AIC->AIC_SVR [AT91C_ID_US1] = (AT91_REG) AT91F_US1Handler;
		AT91C_BASE_AIC->AIC_SMR [AT91C_ID_US1] = 
			(AT91C_AIC_SRCTYPE & AT91C_AIC_SRCTYPE_INT_HIGH_LEVEL) |
			(AT91C_AIC_PRIOR & AT91C_AIC_PRIOR_LOWEST);
		AT91C_BASE_AIC->AIC_ICCR = (1UL << AT91C_ID_US1);		// clear pending interrupt
		AT91C_BASE_AIC->AIC_IECR = (1UL << AT91C_ID_US1);	// enable inerrupt
	}
	else
	{
		// enable the uart
		//         /
		AT91C_BASE_US1->US_CR = AT91C_US_RXEN  | AT91C_US_TXEN;	//    .
	}

#elif CPUSTYLE_ATMEGA

#if CPUSTYLE_ATMEGA128

#elif CPUSTYLE_ATMEGA_XXX4
	// USART initialization
	// Communication Parameters: 8 Data, 1 Stop, No Parity
	// USART Receiver: Off
	// USART Transmitter: Off
	// USART Mode: Asynchronous
	// USART Baud Rate: 9600
	////UCSR0A = 0x00;
	////UCSR0B = 0x00;
	if (withirq)
	{
		UCSR0B |= ((1 << RXCIE0) | (1 << TXCIE0));
	}
	UCSR0C = 0x86;
#else	/* CPUSTYLE_ATMEGA_XXX4 */
	// USART initialization
	// Communication Parameters: 8 Data, 1 Stop, No Parity
	// USART Receiver: Off
	// USART Transmitter: Off
	// USART Mode: Asynchronous
	// USART Baud Rate: 9600
	////UCSRA = 0x00;
	////UCSRB = 0x00;
	if (withirq)
	{
		UCSRB |= ((1 << RXCIE) | (1 << TXCIE));
	}
	UCSRC = 0x86;
#endif	/* CPUSTYLE_ATMEGA_XXX4 */

	// enable pull-up registers for RXD and TXD pins: then rx or tx disabled, these pins disconnected fron UART
	PORTD |= ((1U << PD0) | (1U << PD1));
	DDRD &= ~((1U << DDD0) | (1U << DDD1));

#else
#error Undefined CPUSTYLE_XXX
#endif

}




void 
hardware_cat_set_speed(unsigned baudrate)
{
#if CPUSTYLE_AT91SAM7

	//    
	unsigned value;
	const uint_fast8_t prei = calcdivider_dir(calcdivround(baudrate), AT91SAM7_BRGR_WIDTH, AT91SAM7_BRGR_TAPS, & value);

    AT91C_BASE_US1->US_BRGR = value;
	if (prei == 0)
	{
		AT91C_BASE_US1->US_MR |= AT91C_US_OVER;
	}
	else
	{
		AT91C_BASE_US1->US_MR &= ~ AT91C_US_OVER;
	}

#elif CPUSTYLE_ATMEGA

	//    
	unsigned value;
	const uint_fast8_t prei = calcdivider(calcdivround(baudrate), ATMEGA_UBR_WIDTH, ATMEGA_UBR_TAPS, & value);

#if CPUSTYLE_ATMEGA128
	//if (prei == 0)
	//	UCSR0A |= (1U << U2X0);
	//else
	//	UCSR0A &= ~(1U << U2X0);
	//UBRR0 = value;
#elif CPUSTYLE_ATMEGA_XXX4
	if (prei == 0)
		UCSR0A |= (1U << U2X0);
	else
		UCSR0A &= ~(1U << U2X0);

	UBRR0 = value;
	//UBRR0H = (baudv >> 8) & 0xff;
	//UBRR0L = baudv & 0xff;
#else	/* CPUSTYLE_ATMEGA_XXX4 */
	if (prei == 0)
		UCSRA |= (1U << U2X);
	else
		UCSRA &= ~(1U << U2X);

	UBRRH = (value >> 8) & 0xff;
	UBRRL = value & 0xff;
#endif	/* CPUSTYLE_ATMEGA_XXX4 */



#else
#error Undefined CPUSTYLE_XXX
#endif

}

#if WITHDEBUG	/* serial debug */

/*  ,    */
uint_fast8_t 
hardware_getchar(char * cp)
{
#if CPUSTYLE_AT91SAM7

	if ((AT91C_BASE_US1->US_CSR & AT91C_US_RXRDY) == 0)
		return 0;
	* cp = AT91C_BASE_US1->US_RHR;

#elif CPUSTYLE_ATMEGA32

	#if CPUSTYLE_ATMEGA_XXX4
		if ((UCSR0A & (1 << RX0C)) == 0)
				return 0;
		* cp = UDR0;
	#else	/* CPUSTYLE_ATMEGA_XXX4 */
		if ((UCSRA & (1 << RXC)) == 0)
				return 0;
		* cp = UDR;
	#endif	/* CPUSTYLE_ATMEGA_XXX4 */

#else
#error Undefined CPUSTYLE_XXX
#endif
	return 1;
}

/*      */
uint_fast8_t 
hardware_putchar(uint_fast8_t c)
{

#if CPUSTYLE_AT91SAM7

	if ((AT91C_BASE_US1->US_CSR & AT91C_US_TXRDY) == 0)
		return 0;
	AT91C_BASE_US1->US_THR = c;

#elif CPUSTYLE_ATMEGA32
	#if CPUSTYLE_ATMEGA_XXX4
		if ((UCSR0A & (1 << UDR0E)) == 0)
			return 0;
		UDR0 = c;
	#else	/* CPUSTYLE_ATMEGA_XXX4 */
		if ((UCSRA & (1 << UDRE)) == 0)
			return 0;
		UDR = c;
	#endif	/* CPUSTYLE_ATMEGA_XXX4 */

#else
#error Undefined CPUSTYLE_XXX
#endif
	return 1;
}
#endif

/*      */
void hardware_cat_putchar(uint_fast8_t c)
{
#if CPUSTYLE_AT91SAM7

	AT91C_BASE_US1->US_THR = c;

#elif CPUSTYLE_ATMEGA

	#if CPUSTYLE_ATMEGA128
		//
	#elif CPUSTYLE_ATMEGA_XXX4
		UDR0 = c;
	#else	/* CPUSTYLE_ATMEGA_XXX4 */
		UDR = c;
	#endif	/* CPUSTYLE_ATMEGA_XXX4 */

#else
#error Undefined CPUSTYLE_XXX
#endif
}


#define FIRST_ADC_INPUT 6
#define LAST_ADC_INPUT 7
unsigned char adc_data[LAST_ADC_INPUT-FIRST_ADC_INPUT+1];
#define ADC_VREF_TYPE 0xE0
static unsigned char input_index = 0;
static unsigned char input_ready = 0;
// ADC interrupt service routine
// with auto input scanning
ISR(ADC_vect)
{
	// Read the 8 most significant bits
	// of the AD conversion result
	adc_data[input_index]=ADCH;
	// Select next ADC input
	if (++input_index > (LAST_ADC_INPUT-FIRST_ADC_INPUT))
	{
	   input_index=0;
	   input_ready = 1;
	}
	ADMUX=(FIRST_ADC_INPUT | (ADC_VREF_TYPE & 0xff))+input_index;
	// Delay needed for the stabilization of the ADC input voltage
	_delay_us(10);
	// Start the AD conversion
	ADCSRA |= 0x40;
}

void hardware_adc_initialize(void)
{
	// ADC initialization
	// ADC Clock frequency: 62.500 kHz
	// ADC Voltage Reference: Int., cap. on AREF
	// Only the 8 most significant bits of
	// the AD conversion result are used
	input_index = 0;
	ADMUX = FIRST_ADC_INPUT | (ADC_VREF_TYPE & 0xff);
	_delay_us(10);
	ADCSRA = 0xCF;



}


// Timer 0 overflow interrupt service routine
// 31.250 kHz / 256 = approx 122 Hz
//       
//    .
ISR(TIMER0_OVF_vect)
{
	kbd_spool();	// 
}

//  ,     TICKS_FREQUENCY.
void 
hardware_timer_initialize(void)
{
	// Timer/Counter 0 initialization
	// Clock source: System Clock
	// Clock value: 31.250 kHz
	TCCR0=0x04;
	TCNT0=0x00;

	TIMSK |= 0x01;	// enable interrupt from Timer/Counter 0

}


uint_fast8_t hardware_adc_started(void)
{
	uint_fast8_t f;

	disableIRQ();
	f = input_ready != 0;
	enableIRQ();
	return f;
}

uint_fast8_t 
hardware_ged_pressed_key(uint8_t * scancode)
{
	uint_fast8_t i;
	static volatile unsigned char * const adcsv [] =
	{
		& adc_data [LAST_ADC_INPUT-FIRST_ADC_INPUT - 0],	// ki2
		& adc_data [LAST_ADC_INPUT-FIRST_ADC_INPUT - 1],	// ki1
	};
#define ADCINPUTS (sizeof adcsv / sizeof adcsv [0])
	static const uint_fast8_t mcodes [ADCINPUTS][4] =
	{
		{ 	KBD_CODE_BUTTON1, KBD_CODE_UNUSED, KBD_CODE_UNUSED, KBD_CODE_UNUSED,  },
		{	KBD_CODE_BUTTON2, KBD_CODE_UNUSED, KBD_CODE_UNUSED, KBD_CODE_UNUSED, },
	};

	if (input_ready == 0)
		return 0;
// ideal values: 0, 64, 128, 192
// shifted: 0x20, 0x60, 0xa0, 

	for (i = 0; i < ADCINPUTS; ++ i)
	{
		enum { OFFS = 32 };		
		//    -     
		// ( )   ,    .
		const uint_fast16_t v = (* adcsv [i] & 0xff) + OFFS;
		if ((v & 0x3f) < (OFFS - 4) || (v & 0x3f) > (OFFS + 4))
			continue;
		if (v < 256)
		{
			* scancode = mcodes [i][v / (OFFS * 2)];
			return 1;
		}
	}
	return 0;

}

